#!/usr/bin/env lua

server_version = "Singleton Chat Server v0.1"
server_name = "server"
max_line_length_in = 1024
max_line_length_out = 512
max_channel_count = 60
max_nick_length = 16
max_channel_length = 32
max_user_length = 16
max_topic_length = 160
connect_timeout = 60
ping_interval = 240
ping_timeout = 60
quit_timeout = 60
flood_limit_lines = 100
flood_limit_time = 1
motd = [[

Welcome to the Singleton Chat Server.

]]

moonbridge_io = require "moonbridge_io"

input_sockets = {}
output_sockets = {}
clients = {}
nicks = {}
channels = {}

channel_prototype = {}
channel_metatable = { __index = channel_prototype }

client_prototype = {}
client_metatable = { __index = client_prototype }

do
  local byte = string.byte
  local char = string.char
  local gsub = string.gsub
  local delta_letter = byte("a") - byte("A")
  local function lower_letter(c)
    return char(byte(c) + delta_letter)
  end
  function lower(str)
    str = gsub(str, "[A-Z]", lower_letter)
    str = gsub(str, "%[", "{")
    str = gsub(str, "%]", "}")
    str = gsub(str, "\\", "|")
    str = gsub(str, "~",  "^")
    return str
  end
  local function upper_letter(c)
    return char(byte(c) - delta_letter)
  end
  function upper(str)
    str = gsub(str, "[a-z]", upper_letter)
    str = gsub(str, "{", "[")
    str = gsub(str, "}", "]")
    str = gsub(str, "|", "\\")
    str = gsub(str, "%^",  "~")
    return str
  end
end

local timers, new_timers = {}, {}
function timeout(delay, func)
  local t = {wakeup = moonbridge_io.timeref() + delay, func = func}
  new_timers[t] = true
  return function()
    timers[t] = nil
    new_timers[t] = nil
  end
end

function channel_prototype:send_except(exclude, ...)
  for client in pairs(self.clients) do
    if client ~= exclude then
      client:send(...)
    end
  end
end

function channel_prototype:send(...)
  self:send_except(nil, ...)
end

function client_prototype:init(socket)
  clients[self] = true
  input_sockets[socket] = true
  self.socket = socket
  self.channels = {}
  self.channel_count = 0
  self.read_buffer = {}
  self.read_buffer_len = 0
  self.flood_timer = 0
  self.peer = socket.remote_ip6 or socket.remote_ip4
  --print("Connection from " .. self.peer)  -- DEBUG
  if string.sub(self.peer, 1, 1) == ":" then
    self.peer = "0" .. self.peer
  end
  self.timeout = moonbridge_io.timeref() + connect_timeout
  self.activity = moonbridge_io.timeref()
  if self.connected_hook then
    local errmsg = self:connected_hook()
    if errmsg then
      self:quit(errmsg)
    end
  end
end

function client_prototype:quit(reason)
  if not self.quitting then
    if reason then
      self.socket:flush_nb("ERROR :Connection terminated (" .. reason .. ")\r\n")
    else
      self.socket:flush_nb("ERROR :Connection terminated\r\n")
    end
    self.quitting = true
    self.timeout = moonbridge_io.timeref() + quit_timeout
  end
  output_sockets[self.socket] = true  -- required to gracefully close socket
  if self.registered then
    self:send_all(":", self.prefix, " QUIT :", reason or "No reason")
  end
  if self.nick then
    nicks[lower(self.nick)] = nil
  end
  for channel_name, channel in pairs(self.channels) do
    channel.clients[self] = nil
    if not next(channel.clients) then
      channels[channel_name] = nil
    end
  end
end

function client_prototype:destroy()
  clients[self] = nil
  input_sockets[self.socket] = nil
  output_sockets[self.socket] = nil
end

function client_prototype:abort(reason)
  self.socket:reset()
  self.quitting = true  -- do not send ERROR to client
  self:quit(reason)
  self:destroy()
end

function client_prototype:send_raw(str)
  if self.quitting then return end
  local bytes, errmsg = self.socket:flush_nb(str)
  if not bytes then
    self:abort(errmsg)
  else
    output_sockets[self.socket] = bytes ~= 0
  end
end

function client_prototype:line_to_send(firstarg, ...)
  assert(firstarg ~= nil, "Unexpected nil")
  local chunks
  if type(firstarg) == "number" then
    local cmd = tostring(firstarg)
    while #cmd < 3 do cmd = "0" .. cmd end
    chunks = {":", server_name, " ", cmd, " ", self.nick or "*", " "}
  else
    chunks = { firstarg }
  end
  for i = 1, select("#", ...) do
    local arg = select(i, ...)
    assert(arg ~= nil, "Unexpected nil")
    chunks[#chunks+1] = arg
  end
  return string.sub(table.concat(chunks), 1, max_line_length_out - 2) .. "\r\n"
end

function client_prototype:send(...)
  self:send_raw(self:line_to_send(...))
end

function client_prototype:send_notice(...)
  self:send(":", server_name, " NOTICE ", self.nick or "*", " :", ...)
end

function client_prototype:send_all(...)
  local line = self:line_to_send(...)
  self:send_raw(line)
  local done = { [self] = true }
  for channel_name, channel in pairs(self.channels) do
    for client in pairs(channel.clients) do
      if not done[client] then
        client:send_raw(line)
        done[client] = true
      end
    end
  end
end

function client_prototype:is_command(command, param_count)
  if self.command == command then
    if #self.params >= param_count then
      return true
    end
    self.param_missing = true
  end
  return false
end

function client_prototype:set_prefix()
  self.prefix = self.nick .. "!" .. self.user .. "@" .. self.peer
end

function client_prototype:set_nick(nick)
  if not string.find(nick, "^[A-Za-z[%]\\`_^{|}~0-9-]+$") then
    self:send(432, nick, " :Erroneous nickname")
    return false
  end
  if #nick > max_nick_length then
    self:send(432, nick, " :Erroneous nickname (too long)")
    return false
  end
  local lower_nick = lower(nick)
  local other = nicks[lower_nick]
  if other and other ~= self then
    self:send(433, nick, " :Nickname is already in use")
    return false
  end
  if self.nick_hook then
    local errmsg = self:nick_hook(nick, lower_nick)
    if errmsg then
      if errmsg == true then
        self:send(432, nick, " :Erroneous nickname")
      else
        self:send(432, nick, " :" .. errmsg)
      end
      return false
    end
  end
  if self.lower_nick then
    nicks[self.lower_nick] = nil
  end
  nicks[lower_nick] = self
  self.nick, self.lower_nick = nick, lower_nick
  return true
end

function client_prototype:send_motd()
  self:send(375, ":- ", server_name, " Message of the day - ")
  for line in string.gmatch(motd, "([^\n]*)\n") do
    self:send(372, ":- ", line)
  end
  self:send(376, ":End of MOTD command")
end

function client_prototype:leave(channel_name, lower_channel_name, channel, reason)
  if self.part_hook then
    local errmsg = self:part_hook(channel, channel_name)
    if errmsg then
      self:send_notice("Cannot leave channel ", channel_name, ": ", errmsg)
      return
    end
  end
  channel:send(":", self.prefix, " PART ", channel_name, " :", reason or "No reason")
  self.channels[lower_channel_name] = nil
  self.channel_count = self.channel_count - 1
  channel.clients[self] = nil
  if not next(channel.clients) then
    channels[lower_channel_name] = nil
  end
end

function client_prototype:execute(command, params)
  local old_command       = self.command
  local old_params        = self.params
  local old_param_missing = self.param_missing
  self.command, self.params = command, params
  self.param_missing = false
  if not self.registered then
    if self:is_command("PASS", 1) then
      self._pass = self.params[1]
    elseif self:is_command("NICK", 1) then
      self._nick = self.params[1]
    elseif self:is_command("USER", 4) then
      local user = self.params[1]
      if not string.find(user, "^[0-9A-Za-z_.][0-9A-Za-z_.-]*$") then
        user = "~"
      else
        user = "~" .. string.sub(user, 1, max_user_length)
      end
      self.user = user
      self.real = self.params[4]
    else
      --self:send_notice("Received unsupported command during registration: ", self.command)  -- DEBUG
    end
    if self._nick and self.user then
      if self.auth_hook then
        local errmsg
        errmsg = self:auth_hook(self._nick, self._pass)
        if errmsg then
          self:send(464, ":", errmsg)
          self:quit(errmsg)
          return
        end
      end
      if self:set_nick(self._nick) then
        self._nick = nil
        self._pass = nil
        self:set_prefix()
        self.timeout = nil
        self.nextping = 0
        self.registered = true
        self:send(001, ":Welcome, ", self.prefix)
        self:send(002, ":Your host is \"" .. server_name .. "\", running " .. server_version)
        self:send(003, ":This server was created " .. server_created .. ".")
        self:send(004, server_name, " ", server_version_compact) -- " i n"
        if not self.registered_hook or not self:registered_hook() then
          self:send_motd()
        end
      end
    end
  elseif self.command_hook and self:command_hook() then
    -- do nothing
  elseif self:is_command("PING", 1) then
    local server = self.params[2]
    if server and server ~= server_name then
      self:send(402, server, " :No such server")
    else
      self:send(":", server_name, " PONG ", server_name, " :", self.params[1])
    end
  elseif self.command == "PONG" then
    self.timeout = nil
  elseif self.command == "MOTD" then
    self:send_motd()
  elseif self.command == "QUIT" then
    local reason = self.params[1]
    if not reason or reason == "" then
      reason = "Quit"
    end
    self:quit(reason)
  elseif self:is_command("NICK", 1) then
    if self:set_nick(self.params[1]) then
      self:send_all(":", self.prefix, " NICK :", self.nick)
      self:set_prefix()
    end
  elseif self:is_command("MODE", 1) then
    local target = self.params[1]
    local modes = params[2]
    local lower_target = lower(target)
    if lower_target == self.lower_nick then
      if modes and not modes:match("^[+-i]*$") then
        self:send(501, ":Unknown MODE flag")
      else
        self:send(221, "i")
      end
    elseif channels[lower_target] then
      local extra = params[3]
      if modes then
        if extra then
          self:send(477, target, " :Channel doesn't support modes")
        elseif modes:match("^[+-]?b$") then
          self:send(368, target, " :End of channel ban list")
        elseif modes:match("^[+-]?e$") then
          self:send(349, target, " :End of channel exception list")
        elseif modes:match("^[+-]?I$") then
          self:send(347, target, " :End of channel invite list")
        else
          self:send(477, target, " :Channel doesn't support modes")
        end
      else
        self:send(324, target)
      end
    else
      self:send(403, target, " :No such channel")
    end
  elseif self:is_command("JOIN", 1) then
    if self.params[1] == "0" then
      for channel_name, channel in pairs(self.channels) do
        self:leave(channel_name, channel_name, channel, nil)
      end
    else
      for channel_name in self.params[1]:gmatch("[^,]+") do
        local lower_channel_name = lower(channel_name)
        if not self.channels[lower_channel_name] then
          if self.channel_count >= max_channel_count then
            self:send(405, channel_name, " :You have joined too many channels")
          else
            local channel = channels[lower_channel_name]
            if not channel then
              if not string.find(channel_name, "^[#+&][!\"#$%%&'()+-./0-9;<=>?@A-Z[\\%]^_`a-z{|}~]*$") then
                self:send(403, channel_name, " :No such channel")
              elseif #channel_name > max_channel_length then
                self:send(403, channel_name, " :No such channel (name too long)")
              else
                channel = setmetatable(
                  {
                    lower_name = lower_channel_name,
                    clients = { }
                  },
                  channel_metatable
                )
              end
            end
            if channel then
              local errmsg
              if self.join_hook then
                errmsg = self:join_hook(channel, channel_name)
              end
              if errmsg then
                self:send(403, channel_name, " :No such channel")
                self:send_notice("Cannot join channel ", channel_name, ": ", errmsg)
              else
                channels[lower_channel_name] = channel
                channel.clients[self] = true
                self.channels[lower_channel_name] = channel
                self.channel_count = self.channel_count + 1
                channel:send(":", self.prefix, " JOIN :", channel_name)
                if channel.topic then
                  self:send(332, channel_name, " :", channel.topic)
                else
                  --self:send(331, channel_name, " :No topic is set")
                end
                for client in pairs(channel.clients) do
                  self:send(353, "= ", channel_name, " :", client.nick)
                end
                self:send(366, channel_name, " :End of NAMES list")
              end
            end
          end
        end
      end
    end
  elseif self:is_command("PART", 1) then
    for channel_name in self.params[1]:gmatch("[^,]+") do
      local lower_channel_name = lower(channel_name)
      local channel = self.channels[lower_channel_name]
      if channel then
        self:leave(channel_name, lower_channel_name, channel, self.params[2])
      else
        self:send(442, channel_name, " :You are not on that channel")
      end
    end
  elseif self:is_command("TOPIC", 1) then
    local channel_name = self.params[1]
    local topic = self.params[2]
    local lower_channel_name = lower(channel_name)
    local channel = channels[lower_channel_name]
    if topic then
      self.activity = moonbridge_io.timeref()
    end
    if not channel then
      self:send(403, channel_name, " :No such channel")
    else
      if not topic then
        if channel.topic then
          self:send(332, channel_name, " :", channel.topic)
        else
          self:send(331, channel_name, " :No topic is set")
        end
      else
        topic = topic:sub(1, max_topic_length)
        channel:send(":", self.prefix, " TOPIC ", channel_name, " :", topic)
        if topic == "" then
          channel.topic = nil
        else
          channel.topic = topic
        end
      end
    end
  elseif
    self:is_command("PRIVMSG", 2) or
    self:is_command("NOTICE", 2)
  then
    self.activity = moonbridge_io.timeref()
    local target = self.params[1]
    local message = self.params[2]
    local lower_target = lower(target)
    local client = nicks[lower_target]
    if client then
      client:send(":", self.prefix, " ", self.command, " ", client.nick, " :", message)
    else
      local channel = channels[lower_target]
      if channel then
        channel:send_except(self, ":", self.prefix, " ", self.command, " ", lower_target, " :", message)
      elseif self.command ~= "NOTICE" then
        self:send(401, target, " :No such nick/channel")
      end
    end
  elseif self:is_command("WHO", 0) then
    local target = self.params[1]
    if target then
      local lower_target = lower(target)
      local operflag = self.params[2]
      if operflag ~= "O" and operflag ~= "o" then
        local client = nicks[lower_target]
        if client then
          self:send(352, "* ", client.user, " ", client.peer, " ", server_name, " ", client.nick, " H :0 ", client.real)
        else
          local channel = self.channels[lower_target]
          if channel then
            for client in pairs(channel.clients) do
              self:send(352, target, " ", client.user, " ", client.peer, " ", server_name, " ", client.nick, " H :0 ", client.real)
            end
          end
        end
      end
    end
    self:send(315, target, " :End of WHO list")
  elseif self:is_command("WHOIS", 1) then
    local any_match = false
    local no_match = {}
    local targets = self.params[2] or self.params[1]
    for target in targets:gmatch("[^,]+") do
      local client = nicks[lower(target)]
      if client then
        any_match = true
        local idletime = math.floor(0.5 + moonbridge_io.timeref() - client.activity)
        self:send(311, client.nick, " ", client.user, " ", client.peer, " * :", client.real)
        local line
        for channel_name in pairs(client.channels) do
          if line then
            local line_new = line .. " " .. channel_name
            if #line_new > 56 then
              self:send(319, client.nick, " :", line)
              line = channel_name
            else
              line = line_new
            end
          else
            line = channel_name
          end
        end
        if line then
          self:send(319, client.nick, " :", line)
        end
        self:send(312, client.nick, " ", server_name, " :", server_version)
        self:send(317, client.nick, " ", idletime, " :seconds idle")
      else
        no_match[#no_match+1] = target
      end
    end
    if any_match then
      self:send(318, targets, " :End of WHOIS list")
    end
    for idx, target in ipairs(no_match) do
      self:send(401, target, " :No such nick/channel")
    end
  elseif self:is_command("WHOWAS", 1) then
    local targets = self.params[1]
    for target in targets:gmatch("[^,]+") do
      self:send(406, target, " :There was no such nickname")
    end
    self:send(369, targets, " :End of WHOWAS list")
  elseif
    self.command == "USER" or
    self.command == "PASS"
  then
    self:send(462, ":Unauthorized command (already registered)")
  elseif self.param_missing then
    self:send(461, self.command, " :Not enough parameters")
  else
    self:send(421, self.command, " :Unknown command")
  end
  self.command       = old_command
  self.params        = old_params
  self.param_missing = old_param_missing
end

function client_prototype:process_line(line)
  local now = moonbridge_io.timeref()
  if self.flood_timer < now then
    self.flood_timer = now + flood_limit_time
  else
    self.flood_timer = self.flood_timer + flood_limit_time
  end
  if self.flood_timer - 1e-14 > now + flood_limit_lines * flood_limit_time then
    self:quit("Flooding")
    return
  end
  if not string.find(line, "^ *\r?\n$") then
    local prefix, command, param_string = string.match(
      line,
      "^:([^ \r\n\0]+) +([^ \r\n\0]+) *([^\r\n\0]*)\r?\n$"
    )
    if not prefix then
      command, param_string = string.match(line, "([^ \r\n\0]+) *([^\r\n\0]*)\r?\n$")
    end
    if not command then
      self:send_notice("Received invalid line")
    else
      command = upper(command)
      local first_params, colon, last_param = string.match(param_string, "([^:]*)(:?)(.*)")
      local params = {}
      for param in string.gmatch(first_params, "[^ ]+") do
        params[#params+1] = param
      end
      if colon ~= "" then
        params[#params+1] = last_param
      end
      self:execute(command, params)
    end
  end
end

assert(loadfile((assert(..., "configuration file required as argument"))))(select(2, ...))

server_created = server_created or os.date()
server_version_compact = server_version_compact or string.gsub(server_version, " ", "_")

if uplink then
  uplink.read_buffer = {}
  input_sockets[uplink.socket] = true
  local function handle_write(bytes, errmsg)
    if not bytes then
      error("I/O error with uplink: " .. errmsg)
    end
    if bytes ~= 0 then
      output_sockets[uplink.socket] = bytes ~= 0
    end
  end
  function uplink:write(...)
    return handle_write(uplink.socket:write_nb(...))
  end
  function uplink:flush(...)
    return handle_write(uplink.socket:flush_nb(...))
  end
end
for i, listener in ipairs(listeners) do
  input_sockets[listener] = true
end
local wakeup = nil
while true do
  if moonbridge_io.poll(input_sockets, output_sockets, wakeup) then
    if uplink then
      local bytes, errmsg = uplink.socket:write_nb()
      if not bytes then
        error("I/O error with uplink: " .. errmsg)
      end
      output_sockets[uplink.socket] = bytes ~= 0
      local chunk, status = uplink.socket:read_nb(nil, uplink.terminator)
      if not chunk then
        error("I/O error with uplink: " .. errmsg)
      elseif status == "eof" then
        error("Uplink connection closed")
      elseif chunk ~= "" then
        local buffer = uplink.read_buffer
        local chunk_count = #buffer + 1
        if status ~= "term" then
          buffer[chunk_count] = chunk
        else
          local message
          if chunk_count > 1 then
            buffer[chunk_count] = chunk
            message = table.concat(buffer)
            uplink.read_buffer = {}
          else
            message = chunk
          end
          --io.stdout:write("Received: ", message)  -- DEBUG
          uplink.handler(message)
        end
      end
    end
    for i, listener in ipairs(listeners) do
      local socket, errmsg = listener:accept_nb()
      if socket == nil then
        --print("I/O error when accepting connection: " .. errmsg)  -- DEBUG
      elseif socket then
        if tlsconfig then
          assert(socket:starttls(tlsconfig))
        end
        setmetatable({}, client_metatable):init(socket)
      end
    end
    for client in pairs(clients) do
      local bytes, errmsg = client.socket:write_nb()
      if not bytes then
        client:abort(errmsg)
      else
        local output_pending = bytes ~= 0
        if client.quitting and not output_pending then
          client.socket:close()
          client:destroy()
        else
          output_sockets[client.socket] = output_pending
          local chunk, status = client.socket:read_nb(nil, "\n")
          if not chunk then
            client:abort(status)
          elseif status == "eof" then
            input_sockets[client.socket] = nil
            client:quit("Connection closed")
          elseif chunk ~= "" and not client.quitting then
            local len = client.read_buffer_len + #chunk
            if len > max_line_length_in then
              client:quit("Maximum line length exceeded")
            end
            local buffer = client.read_buffer
            local chunk_count = #buffer + 1
            if status ~= "term" then
              buffer[chunk_count] = chunk
              client.read_buffer_len = len
            else
              local line
              if chunk_count > 1 then
                buffer[chunk_count] = chunk
                line = table.concat(buffer)
                client.read_buffer = {}
              else
                line = chunk
              end
              --io.stdout:write("Received: ", line)  -- DEBUG
              client.read_buffer_len = 0
              client:process_line(line)
            end
          end
        end
      end
    end
  end
  local now = moonbridge_io.timeref()
  wakeup = nil
  for client in pairs(clients) do
    if client.timeout then
      if client.timeout > now then
        local w = client.timeout - now
        if wakeup then wakeup = math.min(wakeup, w)
        else wakeup = w end
      elseif client.quitting then
        client.socket:reset()
        client:destroy()
      elseif client.registered then
        client:quit("Ping timeout")
      else
        client:quit("Timeout during registration")
      end
    elseif client.nextping then
      if client.nextping > now then
        local w = client.nextping - now
        wakeup = math.min(wakeup or w, w)
      else
        client:send("PING :", server_name)
        client.timeout = now + ping_timeout
        client.nextping = now + ping_interval
        if wakeup then wakeup = math.min(wakeup, ping_timeout)
        else wakeup = ping_timeout end
      end
    end
  end
  for timer in pairs(timers) do
    if timer.wakeup <= now then
      timers[timer] = nil
      timer.func()
    end
  end
  for timer in pairs(new_timers) do
    timers[timer] = true
    new_timers[timer] = nil
  end
  for timer in pairs(timers) do
    if wakeup then wakeup = math.min(wakeup, timer.wakeup - now)
    else wakeup = timer.wakeup end
  end
end

